package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** juxtaposition; zero or more adjacent Elements; can specify a rewriting */
public abstract class Sequence extends Element implements Iterable<Element> {

    // Static Constructors //////////////////////////////////////////////////////////////////////////////

    public abstract Sequence and(Sequence s);
    public abstract Sequence not(Sequence s);

    private void needs(Sequence s) { s.needed.add(this); needs.add(s); }
    private void hates(Sequence s) { s.hated.add(this); hates.add(s); }

    void clone(Sequence seq) {
        for(Sequence s : seq.hates) hates(s);
        for(Sequence s : seq.needs) needs(s);
        noFollow = seq.noFollow;
    }

    /** the empty sequence (matches the empty string) */
    public static final Sequence empty = new Sequence.Constant.Empty();

    /** after matching the sequence, do not add anything to the output tree */
    public static Sequence drop(Element[] e, boolean lame) { return new Constant.Drop(e, lame); }

    /** after matching the sequence, insert a constant into the output tree */
    public static Sequence constant(Element[] e, Object o) { return new Constant(e, o); }

    /** after matching the sequence, place the result of the <tt>idx</tt>th match in the output tree */
    public static Sequence singleton(Element[] e, int idx) { return new Singleton(e, idx); }

    /**
     *  after matching the sequence, create the specified output tree
     *  @param tag   the tag for the output tree
     *  @param e     the elements to match
     *  @param drops only elements of <tt>e</tt> whose corresponding <tt>boolean</tt> in <tt>drops</tt> is <i>false</i> will be included in the output tree
     **/
    public static Sequence rewritingSequence(Object tag, Element[] e, boolean[] drops) { return new RewritingSequence(tag, e, drops); }

    ////////////////////////////////////////////////////////////////////////////////

    public Element noFollow = null;
    public String name = null;
    public void setName(String name) { this.name = name; }
    public final Topology noFollow() { return noFollow==null ? null : noFollow.toAtom(); }

    Topology toAtom() {
        if (elements.length!=1) throw new RuntimeException("cannot invoke toAtom() on a Sequence with " + elements.length + " elements: " + this);
        return elements[0].toAtom();
    }

    protected final Element[] elements;

    final HashSet<Sequence> needed = new HashSet<Sequence>();
    final HashSet<Sequence> hated = new HashSet<Sequence>();
    final HashSet<Sequence> needs = new HashSet<Sequence>();
    final HashSet<Sequence> hates = new HashSet<Sequence>();
    public boolean           lame  = false;

    final Position          firstp;
    Position firstp() { return firstp; }

    public Iterator<Element> iterator()    { return new ArrayIterator<Element>(elements); }
    protected Sequence(Element[] elements) {
        this.elements = elements;
        this.firstp = new Position(0);
    }

    // DO NOT MESS WITH THE FOLLOWING LINE!!!
    private Forest.Ref epsilonForm = null;
    private boolean eps = false;
    Forest epsilonForm() {
        if (epsilonForm==null) {
            epsilonForm = new Forest.Ref();
            epsilonForm.merge(firstp().rewrite2(null));
        }
        return epsilonForm;
    }

    protected abstract <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args);


    // Position //////////////////////////////////////////////////////////////////////////////

    /** the imaginary position before or after an element of a sequence; corresponds to an "LR item" */
    public class Position implements IntegerMappable {

        private Forest zero = null;
        public Forest zero() {
            if (zero != null) return zero;
            if (pos > 0) throw new Error();
            return zero = rewrite(null);
        }


                final int      pos;
        private final Position next;
                final Forest[] holder;
        
        private Position(int pos) {
            this.pos      = pos;
            this.next     = pos==elements.length ? null : new Position(pos+1);
            this.holder   = new Forest[elements.length];
        }

        boolean isFirst() { return pos==0; }

        /** the element immediately after this Position, or null if this is the last Position */
        public Element  element() { return pos>=elements.length ? null : elements[pos]; }

        /** the element which produces the sequence to which this Position belongs */
        public Sequence owner() { return Sequence.this; }

        /** the next Position (the Position after <tt>this.element()</tt>) */
        public Position next() { return next; }

        /** true iff this Position is the last one in the sequence */
        public boolean isLast() { return next()==null; }

        // Position /////////////////////////////////////////////////////////////////////////////////

        final <T> Forest<T> rewrite(Input.Location loc) {
            if (this==firstp()) return epsilonForm();
            return rewrite2(loc);
        }

        final <T> Forest<T> rewrite2(Input.Location loc) {
            for(int i=0; i<pos; i++) if (holder[i]==null) throw new Error("realbad " + i);
            for(int i=pos; i<elements.length; i++) {
                if (holder[i]==null) holder[i] = elements[i].epsilonForm();
                if (holder[i]==null) throw new Error("bad " + i);
            }
            Forest<T> ret = Sequence.this.postReduce(loc, holder);
            for(int k=0; k<pos; k++) holder[k] = null; // to help GC
            return ret;
        }

        public String   toString() {
            StringBuffer ret = new StringBuffer();
            ret.append("<{");
            for(Position p = Sequence.this.firstp(); p != null; p = p.next()) {
                ret.append(' ');
                if (p==this) ret.append(" | ");
                if (p.element()!=null) ret.append(p.element());
                else                   ret.append(' ');
            }
            ret.append("}>");
            return ret.toString();
        }
        private final int idx = master_position_idx++;
        public int toInt() { return idx; }
    }
    private static int master_position_idx = 0;

    // toString //////////////////////////////////////////////////////////////////////////////

    public String toString() { return toString(new StringBuffer(), false).toString(); }
    StringBuffer toString(StringBuffer sb) { return toString(sb, true); }
    StringBuffer toString(StringBuffer sb, boolean spacing) {
        for(int i=0; i<elements.length; i++) {
            sb.append(elements[i].toString());
            sb.append(' ');
        }
        return sb;
    }


    // Specialized Subclasses //////////////////////////////////////////////////////////////////////////////

    static class Constant extends Sequence {
        private final Object result;
        public Constant(Element[] e, Object result) { super(e); this.result = result; }
        public Sequence and(Sequence s) { Sequence ret = new Constant(elements, result); ret.needs(s); ret.clone(this); return ret; }
        public Sequence not(Sequence s) { Sequence ret = new Constant(elements, result); ret.hates(s); ret.clone(this); return ret; }
        public <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args) {
            return (Forest<T>)Forest.leaf(loc, result);
        }
        static class Drop extends Constant {
            public Drop(Element[] e, boolean lame) {
                super(e, null);
                this.lame = lame;
            }
            public Sequence and(Sequence s) { Sequence ret = new Drop(elements, lame); ret.needs(s); ret.clone(this); return ret; }
            public Sequence not(Sequence s) { Sequence ret = new Drop(elements, lame); ret.hates(s); ret.clone(this); return ret; }
        }
        static class Empty extends Sequence.Constant.Drop { public Empty() { super(new Element[] { }, false); } }
    }

    static class Singleton extends Sequence {
        private final int idx;
        public Singleton(Element e) { this(new Element[] { e }, 0); }
        public Singleton(Element[] e, int idx) { super(e); this.idx = idx; }
        public <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args) { return (Forest<T>)Forest.singleton(loc, args[idx]); }
        public Sequence and(Sequence s) { Sequence ret = new Singleton(elements, idx); ret.needs(s); ret.clone(this); return ret; }
        public Sequence not(Sequence s) { Sequence ret = new Singleton(elements, idx); ret.hates(s); ret.clone(this); return ret; }
    }

    public static class Unwrap extends Sequence {
        private boolean[] drops;
        public Unwrap(Element[] e)                  { super(e); this.drops = null; }
        public Unwrap(Element[] e, boolean[] drops) { super(e); this.drops = drops; }
        public Sequence and(Sequence s) { Sequence ret = new Unwrap(elements, drops); ret.needs(s); ret.clone(this); return ret; }
        public Sequence not(Sequence s) { Sequence ret = new Unwrap(elements, drops); ret.hates(s); ret.clone(this); return ret; }
        public <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args) {
            for(int i=0; i<args.length; i++) if (args[i]==null) throw new Error();
            if (drops==null) return Forest.create(loc, null, args, true, false);
            int count = 0;
            for(int i=0; i<drops.length; i++) if (!drops[i]) count++;
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            return Forest.create(loc, null, args2, true, false);            
        }
    }

    static class RewritingSequence extends Sequence {
        /*private*/public final Object tag;
        private final boolean[] drops;
        private int count = 0;
        public Sequence and(Sequence s) { Sequence ret = new RewritingSequence(tag, elements, drops); ret.needs(s); ret.clone(this); return ret; }
        public Sequence not(Sequence s) { Sequence ret = new RewritingSequence(tag, elements, drops); ret.hates(s); ret.clone(this); return ret; }
        public RewritingSequence(Object tag, Element[] e) { this(tag, e, null); }
        public RewritingSequence(Object tag, Element[] e, boolean[] drops) {
            super(e);
            this.tag = tag;
            this.drops = drops == null ? new boolean[e.length] : drops;
            for(int i=0; i<this.drops.length; i++) if (!this.drops[i]) count++;
        }
        public <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args) {
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            //System.out.println("reduce \""+tag+"\"");
            return Forest.create(loc, (T)tag, args2, false, false);
        }
        public StringBuffer toString(StringBuffer sb, boolean spacing) {
            int len = sb.length();
            super.toString(sb, spacing);
            len = sb.length()-len;
            if (spacing) for(int i=0; i<50-len; i++) sb.append(' ');
            sb.append(" => ");
            sb.append(tag);
            return sb;
        }
    }
}
